from typing import Optional

# We use the longers paths here as PyCharm will suggest imports from here for other classes
from piate.api import client as client_mod, session as sessions, credentials, resources


class MissingAuthentication(Exception):
    def __init__(self):
        super().__init__(
            "Either a 'session' object or a 'username' and 'api_key' must be provided."
        )


def client(
    username: Optional[str] = None,
    api_key: Optional[str] = None,
    session: Optional[sessions.Session] = None,
):
    has_login = (username is not None) and (api_key is not None)
    has_session = session is None

    if not has_session:
        if has_login:
            session = sessions.Session(
                credentials.Credentials(username=username, api_key=api_key)
            )
        else:
            raise MissingAuthentication()
    return client_mod.Client(session)


__all__ = ["client"]



from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, NewType
from datetime import datetime, timedelta

import requests
from dataclasses_json import dataclass_json, Undefined

from piate.api.version import APIVersion

RefreshToken = NewType("RefreshToken", str)


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class ExpirationBase:
    access: datetime
    refresh: datetime

    def access_expired(self) -> bool:
        return self.access >= datetime.now()

    def refresh_expired(self) -> bool:
        return self.refresh >= datetime.now()


class AuthenticationBase:
    PATH = "/oauth2/token"
    TOKEN_GRANT_TYPE = "password"
    EXTENDS_GRANT_TYPE = "refresh_token"
    API_VERSION: APIVersion

    def _generate_expiration(self) -> Dict:
        now = datetime.now()
        expiration = {
            "access": now + timedelta(hours=3),
            "refresh": now + timedelta(hours=12),
        }
        if self.API_VERSION == APIVersion.V2:
            expiration["id"] = now + timedelta(hours=3)
        return expiration

    @abstractmethod
    def token(self, base_url: str, username: str, password: str) -> Dict:
        response = requests.post(
            f"{base_url}{self.PATH}",
            params={
                "grant_type": self.TOKEN_GRANT_TYPE,
                "username": username,
                "password": password,
            },
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": f"application/vnd.iate.token+json; version={self.API_VERSION.value}",
            },
        ).json()

        response["expiration"] = self._generate_expiration()
        return response

    @abstractmethod
    def extends(self, base_url: str, refresh_token: str) -> Dict:
        response = requests.post(
            f"{base_url}{self.PATH}",
            paarams={
                "grant_type": self.EXTENDS_GRANT_TYPE,
                "refresh_token": refresh_token,
            },
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": f"application/vnd.iate.token+json; version={self.API_VERSION.value}",
            },
        ).json()

        response["expiration"] = self._generate_expiration()
        return response



from dataclasses import dataclass

from piate.api.resources.collections import Collections
from piate.api.resources.domains import Domains
from piate.api.resources.institutions import Institutions
from piate.api.resources.inventories import Inventories
from piate.api.session import Session


@dataclass(init=False)
class Client:
    inventories: Inventories
    collections: Collections
    institutions: Institutions

    def __init__(self, session: Session):
        self._session = session

        self.inventories = Inventories(self._session)
        self.collections = Collections(self._session)
        self.domains = Domains(self._session)
        self.institutions = Institutions(self._session)



from dataclasses import dataclass

import click

import piate
from piate.api.client import Client
from piate.api.credentials import Credentials
from piate.api.session import Session
from piate.cli.format import Format
from piate.cli.render import pages_response_iterator, page_response


@dataclass
class ContextObj:
    client: Client
    format: Format


@click.group()
@click.option(
    "--username",
    "-U",
    metavar="USERNAME",
    help="Username to request against the API",
    required=False,
)
@click.option(
    "--api-key",
    "-K",
    metavar="API_KEY",
    help="API Key to use to request against the API",
    required=False,
)
@click.option("--format", default="json", type=click.Choice(["json", "json-lines"]))
@click.pass_context
def run(ctx: click.Context, username: str, api_key: str, format: str):
    ctx.obj = ContextObj(
        client=Client(Session(Credentials(username=username, api_key=api_key))),
        format=Format(format),
    )


"""
Inventories
"""


@run.group()
def inventories():
    ...


@inventories.command("list-languages")
@click.option(
    "--translation-language", "-L", metavar="TRANSLATION_LANGUAGE", required=False
)
@click.pass_obj
def inventories_languages(obj: ContextObj, translation_language: str):
    kwargs = {}
    if translation_language is not None:
        kwargs["translation_language"] = translation_language
    pages_response_iterator(
        obj.client.inventories.pages_languages(**kwargs), obj.format
    )


@inventories.command("list-query-operators")
@click.option(
    "--translation-language", "-L", metavar="TRANSLATION_LANGUAGE", required=False
)
@click.pass_obj
def inventories_query_operators(obj: ContextObj, translation_language: str):
    kwargs = {}
    if translation_language is not None:
        kwargs["translation_language"] = translation_language
    pages_response_iterator(
        obj.client.inventories.pages_query_operators(**kwargs), obj.format
    )


@inventories.command("list-term-types")
@click.option(
    "--translation-language", "-L", metavar="TRANSLATION_LANGUAGE", required=False
)
@click.pass_obj
def inventories_term_types(obj: ContextObj, translation_language: str):
    kwargs = {}
    if translation_language is not None:
        kwargs["translation_language"] = translation_language
    pages_response_iterator(
        obj.client.inventories.pages_term_types(**kwargs), obj.format
    )


@inventories.command("list-searchable-fields")
@click.option(
    "--translation-language", "-L", metavar="TRANSLATION_LANGUAGE", required=False
)
@click.pass_obj
def inventories_searchable_fields(obj: ContextObj, translation_language: str):
    kwargs = {}
    if translation_language is not None:
        kwargs["translation_language"] = translation_language
    pages_response_iterator(
        obj.client.inventories.pages_searchable_fields(**kwargs), obj.format
    )


@inventories.command("list-primarities")
@click.option(
    "--translation-language", "-L", metavar="TRANSLATION_LANGUAGE", required=False
)
@click.pass_obj
def inventories_primarities(obj: ContextObj, translation_language: str):
    kwargs = {}
    if translation_language is not None:
        kwargs["translation_language"] = translation_language
    pages_response_iterator(
        obj.client.inventories.pages_primarities(**kwargs), obj.format
    )


@inventories.command("list-reliabilities")
@click.option(
    "--translation-language", "-L", metavar="TRANSLATION_LANGUAGE", required=False
)
@click.pass_obj
def inventories_reliabilities(obj: ContextObj, translation_language: str):
    kwargs = {}
    if translation_language is not None:
        kwargs["translation_language"] = translation_language
    pages_response_iterator(
        obj.client.inventories.pages_reliabilities(**kwargs), obj.format
    )


""""
Domains
"""


@run.command("list-domains")
@click.pass_obj
def domains(obj: ContextObj):
    page_response(obj.client.domains.list(), obj.format)


"""
Collections
"""


@run.command("list-collections")
@click.pass_obj
def collections(obj: ContextObj):
    pages_response_iterator(obj.client.collections.pages(), obj.format)


"""
Institutions
"""


@run.command("list-institutions")
@click.pass_obj
def institutions(obj: ContextObj):
    pages_response_iterator(obj.client.institutions.pages(), obj.format)


"""
Entries
"""


@run.group("entries")
def entries():
    ...


@entries.command("search-entries")
def entries_search():
    raise NotImplementedError()


@entries.command("multi-search-entries")
def entries_multi_search():
    raise NotImplementedError()


if __name__ == "__main__":
    run()



from dataclasses import dataclass, field
from typing import Optional, Dict, Type, TypeVar, Generator

from dataclasses_json import dataclass_json, Undefined

from piate.api import API
from piate.api.resources.base import BaseResource
from piate.api.response import create_paged_response_class_generic, Meta

T = TypeVar("T")


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class InventoryElement:
    meta: Meta
    code: str
    name: str
    is_official: Optional[bool] = field(default=None)
    deprecated: Optional[bool] = field(default=None)

    def compact(self) -> Dict:

        compact = {"code": self.code, "name": self.name}
        if self.is_official is not None:
            compact["is_official"] = self.is_official
        if self.deprecated:
            compact["deprecated"] = self.deprecated

        return compact


InventoryElementsResponse = create_paged_response_class_generic(InventoryElement)


class Inventories(BaseResource):
    def pages_languages(
        self, translation_language: Optional[str] = None
    ) -> InventoryElementsResponse:
        return self._pages(
            InventoryElementsResponse,
            "/inventories/_languages",
            translation_language,
        )

    def pages_query_operators(
        self, translation_language: Optional[str] = None
    ) -> InventoryElementsResponse:
        return self._pages(
            InventoryElementsResponse,
            "/inventories/_query-operators",
            translation_language,
        )

    def pages_term_types(
        self, translation_language: Optional[str] = None
    ) -> Generator[InventoryElementsResponse, None, None]:
        return self._pages(
            InventoryElementsResponse,
            "/inventories/_term-types",
            translation_language,
        )

    def pages_searchable_fields(
        self, translation_language: Optional[str] = None
    ) -> Generator[InventoryElementsResponse, None, None]:
        return self._pages(
            InventoryElementsResponse,
            "/inventories/_searchable-fields",
            translation_language,
        )

    def pages_primarities(
        self, translation_language: Optional[str] = None
    ) -> Generator[InventoryElementsResponse, None, None]:
        return self._pages(
            InventoryElementsResponse, "/inventories/_primarities", translation_language
        )

    def pages_reliabilities(
        self, translation_language: Optional[str] = None
    ) -> Generator[InventoryElementsResponse, None, None]:
        return self._pages(
            InventoryElementsResponse,
            "/inventories/_reliabilities",
            translation_language,
        )

    def _pages(
        self, cls: T, path: str, translation_language: Optional[str] = None
    ) -> Generator[T, None, None]:
        params = {"expand": "true", "offset": 0}
        if translation_language is not None:
            params["trans_lang"] = translation_language
        page = cls.from_dict(
            self.session.get(
                path,
                params=params,
                do_auth=False,
                api_type=API.EM,
            )
        )
        yield page

        while page.next is not None:
            page = cls.from_dict(self.session.get_meta_resource(page.next))
            yield page



from dataclasses import dataclass, field
from typing import Optional, Dict

from dataclasses_json import dataclass_json, Undefined

from piate.api.resources.base import BaseResource
from piate.api.response import (
    MetadataType,
    Metadata,
    MetadataResource,
    create_paged_response_class_v2,
)
from piate.api.version import APIVersion


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class CollectionNameLanguage:
    self: MetadataResource
    code: str
    name: str
    is_official: bool


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class CollectionName:
    short_name: str
    full_name: str

    year: Optional[int] = field(default=None)
    language: Optional[CollectionNameLanguage] = field(default=None)

    institution_code: Optional[str] = field(default=None)
    institution_name: Optional[str] = field(default=None)


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class Collection:
    code: str
    id: int
    name: CollectionName
    description: str

    type: MetadataType
    metadata: Metadata
    self: MetadataResource
    update: MetadataResource
    delete: MetadataResource

    def compact(self) -> Dict:
        return dict(
            code=self.code, id=self.id, name=self.name, description=self.description
        )


CollectionPagedResponse = create_paged_response_class_v2(Collection)


class Collections(BaseResource):
    def pages(self):
        page = self._first_page()
        yield page

        while page.next is not None:
            page = CollectionPagedResponse.from_dict(
                self.session.get_metadata_resource(page.next)
            )
            yield page

    def _first_page(self) -> CollectionPagedResponse:
        response = self.session.get(
            "/collections",
            {"expand": "true", "offset": 0, "limit": 10},
            version=APIVersion.V2,
            do_auth=False,
        )
        return CollectionPagedResponse.from_dict(response)



import json
from typing import Optional, Dict, Any, Union

import requests

from piate.api import API
from piate.api.version import APIVersion
from piate.api.authentication.v2 import Authentication, TokenResponse
from piate.api.credentials import Credentials
from piate.api.response import MetadataResource, Meta


class Session:
    BASE_URL: str = "https://iate.europa.eu"

    _credentials: Optional[TokenResponse]
    _http_session: requests.Session

    def __init__(self, credentials: Credentials):
        self._http_session = requests.Session()

        self._oauth_client = Authentication()

        self._username = credentials.username
        self._api_key = credentials.api_key

        self._credentials = None

    @property
    def credentials(self) -> TokenResponse:
        if self._credentials is None:
            self._credentials = self._oauth_client.token(
                base_url=self.BASE_URL, username=self._username, password=self._api_key
            )
        if self._credentials.expiration.access_expired():
            if not self._credentials.expiration.refresh_expired():
                self._credentials = self._oauth_client.extends(
                    base_url=self.BASE_URL,
                    refresh_token=self._credentials.refresh_token,
                )
            else:
                self._credentials = self._oauth_client.token(
                    base_url=self.BASE_URL,
                    username=self._username,
                    password=self._api_key,
                )
        return self._credentials

    def get_metadata_resource(self, resource: MetadataResource) -> Dict:

        api_versions = resource.get_acceptable_api_versions()
        version = api_versions[0]

        response = self._http_session.get(
            resource.href, headers={"Accept": APIVersion.to_mimetype(version)}
        )
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def get_meta_resource(self, resource: Meta) -> Dict:
        response = self._http_session.get(resource.href)
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def get(
        self,
        path: str,
        params: Optional[Dict[str, Union[int, str]]] = None,
        version: Optional[APIVersion] = None,
        api_type: Optional[API] = API.EM,
        do_auth: Optional[bool] = True,
    ) -> Dict:

        if params is None:
            params = {}

        if do_auth:
            raise NotImplementedError()
        else:
            response = self._http_session.get(
                f"{self.BASE_URL}/{api_type.value}-api{path}",
                params=params,
                headers={"Accept": APIVersion.to_mimetype(version)},
            )
            if not response.ok:
                response.raise_for_status()

            return response.json()



from dataclasses import dataclass, field
from typing import TypeVar, List, Generic, Union, Optional

from dataclasses_json import dataclass_json, Undefined

from piate.api.version import APIVersion

R = TypeVar("R")


@dataclass
class Link:
    href: str


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class MetadataResource:
    href: str
    method: str
    accept_media_types: List[str]
    content_media_types: Optional[List[str]] = field(default=None)

    def get_acceptable_api_versions(self) -> List[APIVersion]:
        mimetypes = [APIVersion.from_mimetype(m) for m in self.accept_media_types]
        fitlered_mimetypes = [m for m in mimetypes if m is not None]
        fitlered_mimetypes.sort(reverse=True)
        return fitlered_mimetypes


def create_paged_response_class_v2(item_type):
    @dataclass_json(undefined=Undefined.RAISE)
    @dataclass
    class PagedResponse:
        items: List[item_type]

        offset: int
        limit: int
        size: int

        self: MetadataResource
        create: MetadataResource
        search: MetadataResource
        next: Optional[MetadataResource] = field(default=None)

        def compact(self) -> List[item_type]:
            return self.items

    return PagedResponse


def create_paged_response_class_generic(item_type):
    @dataclass_json(undefined=Undefined.RAISE)
    @dataclass
    class PagedResponse:
        items: List[item_type]

        meta: Meta

        offset: Optional[int] = field(default=None)
        limit: Optional[int] = field(default=None)
        size: Optional[int] = field(default=None)

        next: Optional[Meta] = field(default=None)

        def compact(self) -> List[item_type]:
            return self.items

    return PagedResponse


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class Meta:
    href: str


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class MetadataType:
    self: MetadataResource
    code: Union[int, str]
    name: str


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class MetadataEditTimestamp:
    institution: MetadataType
    timestamp: str
    division: Optional[MetadataType] = None


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class MetadataProtection:
    type: MetadataType
    cascade: Optional[bool] = field(default=None)
    level: Optional[MetadataType] = field(default=None)
    strategy: Optional[MetadataType] = field(default=None)


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class Metadata:
    confidentiality: MetadataType
    creation: MetadataEditTimestamp
    modification: MetadataEditTimestamp
    protection: MetadataProtection



from dataclasses import dataclass, field, asdict
from typing import List, Optional, Dict

from dataclasses_json import Undefined, dataclass_json

from piate.api.resources.base import BaseResource
from piate.api.response import create_paged_response_class_generic
from piate.api.session import Session


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class Domain:
    code: str
    name: str
    level: int
    subdomains: List["Domain"] = field(default_factory=list)

    eurovoc_code: Optional[str] = field(default=None)
    cjeu_code: Optional[str] = field(default=None)
    lookups: Optional[List[str]] = field(default_factory=list)

    def compact(self) -> Dict:
        compact = {
            "code": self.code,
            "name": self.name,
            "level": self.level,
            "subdomains": self.subdomains,
        }

        if self.eurovoc_code:
            compact["eurovoc_code"] = self.eurovoc_code

        if self.cjeu_code:
            compact["cjeu_code"] = self.cjeu_code

        if len(self.lookups):
            compact["lookups"] = self.lookups

        return compact


DomainPagedResponse = create_paged_response_class_generic(Domain)


class Domains(BaseResource):
    def list(self) -> DomainPagedResponse:
        response = self.session.get("/domains/_tree", do_auth=False)
        return DomainPagedResponse.from_dict(response)



from dataclasses import dataclass
from typing import List, Dict, NewType

from dataclasses_json import dataclass_json

from piate.api.authentication.base import (
    AuthenticationBase,
    RefreshToken,
)
from piate.api.version import APIVersion

AuthToken = NewType("AuthToken", str)


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class TokenResponse:
    token_type: str
    tokens: List[AuthToken]
    refresh_token: RefreshToken
    permission_name_by_id: Dict[str:str]
    role_group_name_by_id: Dict[str:str]
    expiration: ExpirationBase


class Authentication(AuthenticationBase):
    API_VERSION = APIVersion.V1

    def token(self, base_url: str, username: str, password: str) -> TokenResponse:
        return TokenResponse.from_dict(
            super().token(base_url=base_url, username=username, password=password)
        )

    def extends(self, base_url: str, refresh_token: str) -> TokenResponse:
        return TokenResponse.from_dict(
            super().extends(base_url=base_url, refresh_token=refresh_token)
        )



import re
from enum import Enum
from typing import Optional


class APIVersion(Enum):
    V1 = "1"
    V2 = "2"

    def __lt__(self, other: "APIVersion") -> bool:
        return int(self.value) < int(other.value)

    @classmethod
    def to_mimetype(cls, version: Optional["APIVersion"] = None) -> str:
        if version is None:
            return "application/json"
        return f"application/vnd.iate.collection+json;version={version.value}"

    @classmethod
    def from_mimetype(cls, mimetype: str) -> Optional["APIVersion"]:
        result = re.search(
            r"application/vnd.iate.collection\+json\;version=(\d+)", mimetype
        )
        if result is not None:
            return APIVersion(result.group(1))
        else:
            return None



from dataclasses import dataclass
from datetime import datetime
from typing import NewType, List

from dataclasses_json import dataclass_json, Undefined

from piate.api.authentication.base import (
    AuthenticationBase,
    RefreshToken,
    ExpirationBase,
)
from piate.api.version import APIVersion

IDToken = NewType("IDToken", str)
AccessToken = NewType("AccessToken", str)


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class Expiration(ExpirationBase):
    id: datetime

    def id_expired(self) -> bool:
        return self.id >= datetime.now()


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class TokenResponseToken:
    id_token: IDToken
    access_token: AccessToken


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class TokenResponse:
    token_type: str
    refresh_token: RefreshToken
    tokens: List[TokenResponseToken]
    expiration: Expiration


class Authentication(AuthenticationBase):
    API_VERSION = APIVersion.V2

    def token(self, base_url: str, username: str, password: str) -> TokenResponse:
        return TokenResponse.from_dict(
            super().token(base_url=base_url, username=username, password=password)
        )

    def extends(self, base_url: str, refresh_token: RefreshToken) -> TokenResponse:
        return TokenResponse.from_dict(
            super().extends(base_url=base_url, refresh_token=refresh_token)
        )



from dataclasses import dataclass, field
from typing import Optional, Dict

from dataclasses_json import dataclass_json, Undefined

from piate.api import API
from piate.api.resources.base import BaseResource
from piate.api.response import create_paged_response_class_generic, Meta


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class InstitutionMetadata:
    creation_date: str
    change_date: str


@dataclass_json(undefined=Undefined.RAISE)
@dataclass
class Institution:
    meta: Meta
    code: str
    name: str
    metadata: InstitutionMetadata
    parent: Optional[str] = field(default=None)

    def compact(self) -> Dict:
        compact = {"code": self.code, "name": self.name, "metadata": self.metadata}
        if self.parent:
            compact["parent"] = self.parent

        return compact


InstitutionPagedResponse = create_paged_response_class_generic(Institution)


class Institutions(BaseResource):
    def pages(self) -> InstitutionPagedResponse:
        page = self._first_page()
        yield page

        while page.next is not None:
            page = InstitutionPagedResponse.from_dict(
                self.session.get_meta_resource(page.next)
            )
            yield page

    def _first_page(self) -> InstitutionPagedResponse:
        response = self.session.get(
            "/institutions",
            params={"expand": "true", "offset": 0},
            do_auth=False,
            api_type=API.UAC,
        )
        return InstitutionPagedResponse.from_dict(response)



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="piate",
    version="0.1.4",
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    url="https://www.github.com/drewsonne/piate",
    license="Apache",
    author="Drew J. Sonne",
    author_email="drew.sonne@gmail.com",
    description="Python Library to interact with the IATE database",
    long_description=README,
    long_description_content_type="text/markdown",
    install_requires=["requests", "dataclasses-json", "click"],
    entry_points={"console_scripts": ["iate=piate.cli:run"]},
)



import dataclasses
import json
from functools import reduce
from json import JSONEncoder
from typing import Any, List, Iterator

import click

from piate.cli.format import Format


class _IATEResponseEncoder(JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            if hasattr(o, "items"):
                return [i.compact() for i in o.items]
            elif hasattr(o, "compact"):
                return o.compact()
            else:
                return dataclasses.asdict(o)
        else:
            return super().default(o)


def response(obj: Any):
    click.echo(json.dumps(obj, cls=_IATEResponseEncoder, indent=4))


def response_lines(obj: List[Any]):
    for item in obj:
        click.echo(json.dumps(item, cls=_IATEResponseEncoder))


def pages_response_iterator(paginator: Iterator, format: Format):
    if format == Format.JSON:
        response(list(reduce(lambda items, page: items + page.items, paginator, [])))
    elif format == Format.JSON_LINES:
        for page in paginator:
            response_lines(page.items)
    else:
        raise RuntimeError(f"Unexpected output format: {format=}")


def page_response(page, format: Format):
    if format == Format.JSON:
        response(page)
    elif format == Format.JSON_LINES:
        response_lines(page.items)
    else:
        raise RuntimeError(f"Unexpected output format: {format=}")
